/*
 * jProlog.java
 *
 * Copyright(c) 2002, Particle Corp.
 */

import java.io.IOException;
import java.io.StringReader;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Stack;

/**
 * jProlog
 *
 * "Prolog accepts facts and rules as a set of axioms, and the user's
 * question as a conjectured theorem; then it tries to prove this
 * theorem - that is, to demonstrate that it can be logically
 * derived from the axioms." - Ivan Bratko
 *
 * @author Alex S.
 * @version 0.0.1
 */
public class jProlog {

    /**
     * this is the database for all our clauses
     */
    private static Hashtable database = new Hashtable();

    private static BufferedReader r;

    public static void assertRule(jPrologRuleInterface rule) throws jPrologException {
        jPrologNode n = rule.getMatcher();
        String name = n.getName() + "/" + n.getArity();
        if(n.getName().equals(":-"))   // if rule
            name = n.getHead().getName() + "/" + n.getHead().getArity();
        n.setFunction(rule);
        database.put(name,new jPrologList(n,null));
    }

    static {
        try{
            assertRule(new jPrologRule_consult());
            assertRule(new jPrologRule_listing());

            // place to add more stuff

        }catch(jPrologException e){
            //..
        }

    }


    /**
     * display our parse tree...
     */
    public static void executeQuery(jPrologNode root) throws IOException, jPrologException {

        Hashtable h = new Hashtable();
        jPrologResolver resolver = new jPrologResolver(root,database);
        boolean moreData = true;
        while(moreData && resolver.resolve()){
            h = resolver.getVariables();
            if(h.size() == 0){
                System.out.println("yes");
                moreData = false;
            }else{
                StringBuffer sb = new StringBuffer();
                Enumeration enum = h.keys();
                while(enum.hasMoreElements()){
                    String name = (String)enum.nextElement();
                    jPrologNode value = (jPrologNode)h.get(name);
                    sb.append(name + " = " + value + "\n");
                }
                sb.setCharAt(sb.length()-1,'?');
                System.out.print(sb.toString());

                String l = r.readLine();
                if(!l.startsWith(";"))
                    moreData = false;
            }
        }
        if(moreData)
            System.out.println("no");
    }

    /**
     * entry point to our app - allows user to type in various commands
     *
     * @param args Command line arguments
     */
    public static void main(String[] args) throws IOException, jPrologException {

        System.out.println("pProlog - POC 0.0.7a");
        System.out.println("Copyright(c) 2002, Particle Corp.\n");

        String prompt = "| ?- ";
        r = new BufferedReader(new InputStreamReader(System.in));
        System.out.print(prompt);
        String line = r.readLine();
        while(line != null){
            line = line.trim();
            if(line.startsWith("halt") || line.startsWith("exit") || line.startsWith("quit"))
                break;
            try{
                if(line.length() > 0)
                    executeQuery(jPrologParser.parse(line));
            }catch(jPrologException e){
                System.out.println("*** "+e.getMessage());
            }catch(Exception e){
                e.printStackTrace();
            }
            System.out.print(prompt);
            line = r.readLine();
        }
    }
}

