/*
 * jPrologLexer.java
 *
 * Copyright(c) 2002, Particle Corp.
 */

import java.io.IOException;
import java.io.Reader;
import java.io.LineNumberReader;

/**
 * Simple lexer for jProlog
 * (not very advanced, but functional...)
 *
 * @author Alex S.
 * @version 0.0.1
 */
public class jPrologLexer {

    // buffered reader
    private LineNumberReader reader;

    public static final int ATOM = 1;
    public static final int SPECIAL = 2;
    public static final int EOF = -1;

    // current token type
    private int currentType;
    private StringBuffer currentToken;

    /**
     * Initializes the lexer to read from the reader
     *
     * @param r Reader from which to grab raw data.
     */
    public jPrologLexer(Reader r) {
        reader = new LineNumberReader(r);
        currentType = 0;
        currentToken = new StringBuffer();
    }

    /**
     * gets token type
     */
    public int getTokenType(){
        return currentType;
    }

    /**
     * gets token as string
     */
    public String getToken(){
        return currentToken.toString();
    }

    /**
     * gets current line number
     */
    public int getLineNumber(){
        return reader.getLineNumber();
    }

    /**
     * method that reads all the fluf (spaces/comments) and gets to the
     * first important character
     */
    private int fluf() throws IOException {
        int ch;
        for(;;){
            ch = reader.read();
            if(ch < 0)
                break;
            while(Character.isWhitespace((char)ch))
                  ch = reader.read();
            if(ch == '%'){  // comment
                ch = reader.read();
                while(ch != '\n' && ch >= 0)
                    ch = reader.read();
            }else
                break;
        }
        return ch;
    }

    /**
     * grabs a prolog token from stream; returns all tokens as Strings
     */
    public String nextToken() throws IOException {
        currentToken.setLength(0);
        if(currentType == EOF)
            return currentToken.toString();
        int ch =  fluf();
        if(ch < 0){
            currentType = EOF;
            return currentToken.toString();
        }
        if(ch == '[' || ch == ']' || ch == '(' || ch == ')' || ch == ',' || ch == ';' || ch == '|'){
            currentToken.append((char)ch);
            currentType = SPECIAL;
            return currentToken.toString();
        }

        if(Character.isLetter((char)ch) || ch == '_'){
            do{
                currentToken.append((char)ch);
                reader.mark(7);
                ch = reader.read();
            }while(ch >= 0 && ( Character.isLetter((char)ch) || Character.isDigit((char)ch) || ch == '_'));
            reader.reset();
        }else if(Character.isDigit((char)ch)){
            do{
                currentToken.append((char)ch);
                reader.mark(7);
                ch = reader.read();
            }while(ch >= 0 && Character.isDigit((char)ch));
            reader.reset();
        }else if(ch == '\''){
            ch = reader.read();
            while(ch >= 0 && ch != '\''){
                if(ch == '\\'){
                    ch = reader.read();
                    if(ch < 0)
                        break;
                }
                currentToken.append((char)ch);
                ch = reader.read();
            }
        }else if(ch == '+' || ch == '-' || ch == '*' || ch == '/' || ch == '<' || ch == '>' || ch == '=' ||
                 ch == ':' || ch == '.' || ch == '&' || ch == '_' || ch == '~' || ch == '!'){
            do{
                currentToken.append((char)ch);
                reader.mark(7);
                ch = reader.read();
            }while(ch >= 0 && (ch == '+' || ch == '-' || ch == '*' || ch == '/' || ch == '<' ||
                               ch == '>' || ch == '=' || ch == ':' || ch == '.' || ch == '&' ||
                               ch == '_' || ch == '~' || ch == '!'));
            reader.reset();
        }
        currentType = ATOM;
        return currentToken.toString();
    }

}

