/*
 * jPrologMatcher.java
 *
 * Copyright(c) 2002, Particle Corp.
 */

import java.io.IOException;
import java.io.StringReader;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Stack;

/**
 * jProlog Matcher
 *
 * Handles prolog matching (unification)
 *
 * @author Alex S.
 * @version 0.0.1
 */
public class jPrologMatcher {

    // have a singleton instance
    private static jPrologMatcher singleton = null;

    /**
     * prevent anyone from creating an instance
     */
    private jPrologMatcher(){
        //...
    }

    /**
     * returns our singleton instance
     */
    public static jPrologMatcher getMatcher(){
        if(singleton == null)
            singleton = new jPrologMatcher();
        return singleton;
    }

    /**
     * construct a variable from other temporary variables.
     */
    public jPrologNode constructVariable(jPrologNode n,Hashtable h){
        if(n == null)
            return null;
        jPrologNode o = n;
        while(n != null && n.isVariable()){
            o = n;
            n = (jPrologNode)h.get(n.getName());
        }
        if(n == null)
            n = o;
        return n;
    }

    /**
     * given 2 terms, they match if they're identical, or the variables
     * in them can be instantiated in such a way that they become identical.
     *
     * @param a Node1
     * @param b Node2
     * @param h Instantiated Variables are returned as part of this hash.
     * @return true if match is found, false otherwise.
     */
    public boolean match(jPrologNode a,jPrologNode b,Hashtable h){
        if(a == null && b == null)
            return true;    // they match
        if(a == null || b == null)
            return false;   // they don't match

        if(a.isVariable() || b.isVariable()){
            jPrologNode anext = a.getNext();
            jPrologNode bnext = b.getNext();
            a = constructVariable(a,h);
            b = constructVariable(b,h);
            if(a.isVariable() && b.isVariable() && !a.getName().equals(b.getName())){
                h.put(a.getName(),b.cloneNoNext());
                return match(anext,bnext,h);
            }else if(a.isVariable()){   // a is a variable
                h.put(a.getName(),b.cloneNoNext());
                return match(anext,bnext,h);
            }else if(b.isVariable()){   // b is a variable,
                h.put(b.getName(),a.cloneNoNext());
                return match(anext,bnext,h);
            }else{  // both are not variables
                return match(a.cloneNoNext(),b.cloneNoNext(),h) && match(anext,bnext,h);
            }
        }

        // if term's names are not the same
        if(!a.getName().equals(b.getName()))
            return false;

        // their arity is different
        if(a.getArity() != b.getArity())
            return false;

        // do their children match?
        return match(a.getHead(),b.getHead(),h) &&
            match(a.getNext(),b.getNext(),h);
    }


}

