/*
 * jPrologNode.java
 *
 * Copyright(c) 2002, Particle Corp.
 */

import java.util.Vector;

/**
 * jProlog node class. This node can represent atoms, numbers, composite
 * terms, facts, rules, and directives.
 *
 * NOTE: we're ignoring directives for the time being.
 *
 * @author Alex S.
 * @version 0.0.1
 */
public class jPrologNode {

    /**
     * this name is either an atom, number, variable name, or functor name.
     */
    protected String name;

    /**
     * parse tree is just a binary tree
     */
    protected jPrologNode head, next;

    /**
     * functionality behind this thing.
     */
    protected jPrologRuleInterface function;

    /**
     * construct a node by this name
     */
    public jPrologNode(String s){
        name = s;
        head = null;
        next = null;
        function = null;
    }

    public void setFunction(jPrologRuleInterface f){
        function = f;
    }
    public jPrologRuleInterface getFunction(){
        return function;
    }

    public void setHead(jPrologNode h){
        head = h;
    }
    public jPrologNode getHead(){
        return head;
    }

    public void setNext(jPrologNode n){
        next = n;
    }
    public jPrologNode getNext(){
        return next;
    }

    public void setName(String s){
        name = s;
    }
    public String getName(){
        return name;
    }

    public int getArity(){
        if(name.equals("."))    // if it's a list
            return 2;
        return countList(head);
    }

    private int countList(jPrologNode n){
        if(n == null)
            return 0;
        return 1 + countList(n.next);
    }

    public boolean isVariable(){
        if(head == null && (name.charAt(0) == '_' || Character.isUpperCase(name.charAt(0))))
            return true;
        return false;
    }

    /**
     * clones the node without copying the 'next' node.
     */
    public jPrologNode cloneNoNext() {
        if(next == null)
            return this;
        jPrologNode o = new jPrologNode(name);
        o.setHead(head);
        o.function = function;
        return o;
    }


    /**
     * convert a prolog list to a vector
     */
    public Vector toList(){
        Vector v = new Vector();
        jPrologNode h = this;
        while(h.getName().equals(".") || !h.getName().equals("[]")){
            v.addElement(h.getHead().cloneNoNext());
            h = h.getHead().getNext();
        }
        if(!h.getName().equals("[]"))
            v.addElement(h);
        return v;
    }

    /**
     * print out a list
     */
    public String strList(){
        StringBuffer sb = new StringBuffer();
        sb.append("[");
        jPrologNode h = this;
        while(h.getName().equals(".")){
            sb.append(h.getHead().cloneNoNext() + ",");
            h = h.getHead().getNext();
        }
        if(!h.getName().equals("[]")){
            sb.setCharAt(sb.length()-1,'|');
            sb.append(h+",");
        }
        sb.setCharAt(sb.length()-1,']');
        return sb.toString();
    }

    /**
     * create a string implementation of this thing.
     */
    public String toString(){
        StringBuffer sb = new StringBuffer();
        if(name.equals(".")){
            sb.append(strList());
        }else if(name.equals(":-")){
            sb.append(getHead().cloneNoNext() + " :- " + getHead().getNext());
        }else{
            sb.append(name);
            if(head != null)
                sb.append("(" + head + ")");
        }
        if(next != null)
            sb.append("," + next);
        return sb.toString();
    }
}

