/*
 * jPrologParser.java
 *
 * Copyright(c) 2002, Particle Corp.
 */

import java.io.IOException;
import java.io.StringReader;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * simple parser for jProlog
 *  - does not handle operators...
 *
 * @author Alex S.
 * @version 0.0.2
 */
public class jPrologParser {

    // the lexer
    private jPrologLexer lexer;

    /**
     * construct a new prolog parser
     */
    public jPrologParser(jPrologLexer l){
        lexer = l;
    }

    /**
     * do a quick parse of a single clause
     *
     * @param s String which to parse.
     * @return the single clause that was parsed.
     */
    public static jPrologNode parse(String s) throws IOException, jPrologException {
        StringReader sr = new StringReader(s);
        jPrologLexer lexer = new jPrologLexer(sr);
        jPrologParser parser = new jPrologParser(lexer);
        Vector clauses = parser.parse();
        if(clauses.size() != 1)
            throw new jPrologException("quickparse is only capable of 1 clause");
        return (jPrologNode)clauses.firstElement();
    }

    /**
     * parse a prolog program
     */
    public Vector parse() throws IOException, jPrologException {
        lexer.nextToken();
        Vector v = new Vector();
        jPrologNode term = term();
        while(term != null){
            v.addElement(term);
            term = term();
        }
        // if parser has not reached the end...
        if(lexer.getTokenType() != jPrologLexer.EOF)
            throw new jPrologParserException("could not reach end",lexer.getLineNumber());
        if(v.isEmpty())
            throw new jPrologParserException("nothing parsable",lexer.getLineNumber());
        return v;
    }

    /**
     * ll(1) matcher
     */
    private boolean match(String s) throws IOException {
        if(lexer.getToken().equals(s)){
            lexer.nextToken();
            return true;
        }
        return false;
    }

    /**
     * takes named lists, and makes ordinary lists out of them.
     */
    private jPrologNode makeList(jPrologNode n,jPrologNode t){
        if(n == null)
            return t;
        jPrologNode p = new jPrologNode(".");
        p.setHead(n);
        n.setNext(makeList(n.getNext(),t));
        return p;
    }

    /**
     * parses a term
     */
    private jPrologNode term() throws IOException, jPrologException {
        jPrologNode n = null;
        if(match("[")){
            n = term();
            jPrologNode q = new jPrologNode("[]");
            if(match("|")){
                q = term();
                if(q == null)
                    throw new jPrologParserException("term expected after '|'",lexer.getLineNumber());
                if(q.getNext() != null)
                    throw new jPrologParserException("only 1 term expected after '|'",lexer.getLineNumber());
            }
            n = makeList(n,q);
            if(!match("]"))
                throw new jPrologParserException("']' expected",lexer.getLineNumber());
        }else if(lexer.getTokenType() == jPrologLexer.ATOM){  // functor
            n = new jPrologNode(lexer.getToken());
            lexer.nextToken();
            if(match("(")){
                n.setHead(term());
                if(!match(")"))
                    throw new jPrologParserException("')' expected",lexer.getLineNumber());
            }
        }

        // parser hack to allow regular Prolog rules
        if(match(":-")){
            jPrologNode q = n;
            n = term();
            n = makeList(n,new jPrologNode("[]"));
            q.setNext(n);
            n = new jPrologNode(":-");
            n.setHead(q);
        }else if(match(","))
            n.setNext(term());
        match(".");     // match end with '.'
        return n;
    }
}

